#include <WiFi.h>

// NTP settings
#define TIMEZONE 9 // timezone (GMT = 0, Japan = 9)
#define NTP_SERVER "pool.ntp.org"

// WiFi settings
#define WIFI_SMARTCONFIG false

#if !WIFI_SMARTCONFIG
// if you do not use smartConfifg, please specify SSID and password here
#define WIFI_SSID "SSID" // your WiFi's SSID
#define WIFI_PASS "PASS" // your WiFi's password
#endif

// clock mode settings : show seconds when you pick up the watch to confirm whether the watch shows correct time
// showing seconds mode : approx. 200 turns per hour (4800 turns a day). Too many for most wrist watches.
// showing minutes mode : approx. 15 turns per hour (360 turns a day). Too few for most wrist watches.
// By combining two mode, you can adjust the number of turns per day.
// ex. ETA7750 based movement needs 1600 (800turns for CW rotation), but ETA2892 needs only 650.
#define SHOW_SECOND_START 6 // turn on at 6 o'clock
#define SHOW_SECOND_STOP 9 // turn off at 9 o'clock

//////////////////////////////
// motor control section
//////////////////////////////

// Motor parameters
#define STEPS_PER_ROTATION 2048 // steps of a single rotation of the motor
#define POSITION 60 // each hand has 60 positions
#define MOTOR_DELAY 2400
#define INITIAL_POS 45 // freeplay of the stepper motor

#define PHASE 4
#define PINS 4
// ports used to control the stepper motor.
// if your motor rotates to the opposite direction, change to
//int port[PINS] = {10, 8, 7, 6};
int port[PINS] = {6, 7, 8, 10};

// sequence of stepper motor control
int seq[PHASE][PINS] = {
  {  LOW,  LOW, HIGH,  LOW},
  {  LOW,  LOW,  LOW, HIGH},
  { HIGH,  LOW,  LOW,  LOW},
  {  LOW, HIGH,  LOW,  LOW}
};

// stepper motor rotation
void rotate(int step) {
  static int phase = 0;
  int i, j;
  int delta = (step > 0) ? 1 : (PHASE-1);
  int contmode = 0;
  
  if(abs(step) <= STEPS_PER_ROTATION / POSITION + 1) { // 1 sec advance
    contmode = 1;
  }
  
  step = (step > 0) ? step : -step;
  for(j = 0; j < step; j++) {
    phase = (phase + delta) % PHASE;
    for(i = 0; i < PINS; i++) {
      digitalWrite(port[i], seq[phase][i]);
    }
    if(contmode) {
      delayMicroseconds(1000000L / (STEPS_PER_ROTATION / POSITION + 1));
    }
    else {
      delayMicroseconds(MOTOR_DELAY);
      if(j < 20) delayMicroseconds(MOTOR_DELAY); // acceleration
      if(j < 10) delayMicroseconds(MOTOR_DELAY); // acceleration
      if(step - j < 20) delayMicroseconds(MOTOR_DELAY*2); // deacceleration
      if(step - j < 10) delayMicroseconds(MOTOR_DELAY*2); // deacceleration
      if(step - j < 5) delayMicroseconds(MOTOR_DELAY*2); // deacceleration
    }
  }
  // power cut
  for(i = 0; i < PINS; i++) {
    digitalWrite(port[i], LOW);
  }
}

//////////////////////////////
// rorary clock control section
//////////////////////////////

#define DIGIT 3
typedef struct {
  int v[DIGIT];
} Digit;

void printDigit(Digit d);
Digit rotUp(Digit current, int digit, int num);
Digit rotDown(Digit current, int digit, int num);
Digit rotDigit(Digit current, int digit, int num);
Digit setDigit(Digit current, int digit, int num);
int setNumber(Digit n);

// avoid error accumuration of fractional part of 4096 / POSITION
void rotStep(int s) {
  static long currentPos;
  static long currentStep;
  
  currentPos += s;
  long diff = currentPos * STEPS_PER_ROTATION / POSITION - currentStep;
  rotate(diff);
  currentStep += diff;
}

void printDigit(Digit d) {
  String s = "            ";
  int i;

  for(i = 0; i < DIGIT; i++) {
    s.setCharAt(i, d.v[i] + '0');
  }
  Serial.println(s);
}

Digit current = {0, 0, 1};

//increase specified digit
Digit rotUp(Digit current, int digit, int num) {
  int freeplay = 0;
  int i;
  
  for(i = digit; i < DIGIT - 1; i++) {
    int id = current.v[i];
    int nd = current.v[i+1];
    if(id <= nd) id += POSITION;
    freeplay += id - nd - 1;
  }
  freeplay += num;
  rotStep(-1 * freeplay);
  current.v[digit] = (current.v[digit] + num) % POSITION;
  for(i = digit + 1; i < DIGIT; i++) {
    current.v[i] = (current.v[i - 1] + (POSITION - 1)) % POSITION;
  }

  Serial.print("up end : ");
  printDigit(current);
  return current;
}

// decrease specified digit
Digit rotDown(Digit current, int digit, int num) {
  int freeplay = 0;
  int i;
  
  for(i = digit; i < DIGIT - 1; i++) {
    int id = current.v[i];
    int nd = current.v[i+1];
    if(id > nd) nd += POSITION;
    freeplay += nd - id;
  }
  freeplay += num;
  rotStep( 1 * freeplay);
  current.v[digit] = (current.v[digit] - num + POSITION) % POSITION;
  for(i = digit + 1; i < DIGIT; i++) {
    current.v[i] = current.v[i - 1];
  }

  Serial.print("down end : ");
  printDigit(current);
  return current;
}

// decrease or increase specified digit 
Digit rotDigit(Digit current, int digit, int num) {
  if(num > 0) {
    return rotUp(current, digit, num);
  }
  else if(num < 0) {
    return rotDown(current, digit, -num);
  }
  else return current;
}

// set single digit to the specified number
Digit setDigit(Digit current, int digit, int num) {
  int pd, cd = current.v[digit];
  if(digit == 0) { // most significant digit
    pd = 0;
  }
  else {
    pd = current.v[digit - 1];
  }
  if(cd == num) return current;
  
  // check if increasing rotation is possible
  int n2 = num;
  if(n2 < cd) n2 += POSITION;
  if(pd < cd) pd += POSITION;
  if(pd <= cd || pd > n2) {
    return rotDigit(current, digit, n2 - cd);
  }
  // if not, do decrease rotation
  if(num > cd) cd += POSITION;
  return rotDigit(current, digit, num - cd);  
}

int setNumber(Digit n) {
  int i;
  
  for(i = 0; i < DIGIT; i++) {
    if(current.v[i] != n.v[i]) {
      break;
    }
  }
  // rotate most significant wheel only to follow current time ASAP
  if(i < DIGIT) {
      current = setDigit(current, i, n.v[i]);
  }
  if(i >= DIGIT - 1) {
    return true; // finished
  }
  else {
    return false;
  }
}


//////////////////////////////
// WiFi and NTP section
//////////////////////////////

// user interface
#include <Adafruit_NeoPixel.h>
#define LED_PIN 2
Adafruit_NeoPixel pixels(1, LED_PIN, NEO_GRB + NEO_KHZ800);
#define LED_BRIGHTNESS 128

#define BUTTON_PIN 3
#define LONG_PRESS_TIME 3000

void getNTP(void) {
  for(int i = 0; WiFi.status() != WL_CONNECTED; i++) {
    if(i > 30) {
      ESP.restart();
    }
    Serial.println("Waiting for WiFi connection..");
    delay(1000);
  }

  configTime(TIMEZONE * 3600L, 0, NTP_SERVER);
  printLocalTime();
}

void printLocalTime() {
  struct tm timeinfo;
  if (!getLocalTime(&timeinfo)) {
    Serial.println("Failed to obtain time");
    return;
  }
}

void wifiSetup() {
  WiFi.mode(WIFI_STA);
#if WIFI_SMARTCONFIG
  WiFi.begin();
#else
  WiFi.begin(WIFI_SSID, WIFI_PASS);
#endif

  for (int i = 0; ; i++) {
    Serial.println("Connecting to WiFi...");
    delay(1000);
    if (WiFi.status() == WL_CONNECTED) {
      break;
    }
#if WIFI_SMARTCONFIG
  if(i > 30)
    break;
#endif    
  }

#if WIFI_SMARTCONFIG
  if (WiFi.status() != WL_CONNECTED) {
    WiFi.mode(WIFI_AP_STA);
    WiFi.beginSmartConfig();

    //Wait for SmartConfig packet from mobile
    Serial.println("Waiting for SmartConfig.");
    while (!WiFi.smartConfigDone()) {
      Serial.print(".");
      pixels.setPixelColor(0, pixels.Color(LED_BRIGHTNESS, LED_BRIGHTNESS, LED_BRIGHTNESS)); // white
      pixels.show();
      delay(1000);
    }

    Serial.println("");
    Serial.println("SmartConfig received.");

    //Wait for WiFi to connect to AP
    Serial.println("Waiting for WiFi");
    while (WiFi.status() != WL_CONNECTED) {
      delay(1000);
      Serial.print(",");
    }
  }
  Serial.println("WiFi Connected.");
#endif

  Serial.print("SSID: ");
  Serial.println(WiFi.SSID());
  Serial.print("IP Address: ");
  Serial.println(WiFi.localIP());
}

//////////////////////////////
// setup and main loop
//////////////////////////////

void setup() {
  Serial.begin(115200);
  Serial.println("start");

  pinMode(port[0], OUTPUT);
  pinMode(port[1], OUTPUT);
  pinMode(port[2], OUTPUT);
  pinMode(port[3], OUTPUT);

  pixels.setPixelColor(0, pixels.Color(LED_BRIGHTNESS, LED_BRIGHTNESS, 0)); // yellow
  pixels.show();

  wifiSetup();
  
  pixels.setPixelColor(0, pixels.Color(LED_BRIGHTNESS, 0, LED_BRIGHTNESS)); // purple
  pixels.show();

  rotate(STEPS_PER_ROTATION * DIGIT); // reset all digits using physical end stop

  rotate(-INITIAL_POS); // compensate the free play

  getNTP(); // get current time

  pinMode(BUTTON_PIN, INPUT_PULLUP);
}

boolean show_second = false;
boolean turn_off = false;

void loop() {
  static int prevhour = -1;
  struct tm tmtime;
  boolean finished;
  Digit n;

  getLocalTime(&tmtime);

  // turn on/off of showing seconds automatically 
  if(tmtime.tm_hour == SHOW_SECOND_START && tmtime.tm_min == 0) {
    show_second = true;
  }
  else if(tmtime.tm_hour == SHOW_SECOND_STOP && tmtime.tm_min == 0) {
    show_second = false;
  }

  // manual operation
  if(digitalRead(BUTTON_PIN) == LOW) {
    unsigned long start = millis();
    boolean long_press = false;
    while(digitalRead(BUTTON_PIN) == LOW) {
      if(millis() > start + LONG_PRESS_TIME) {
        if(turn_off) {
          ESP.restart(); // reset
        }
        else {
          pixels.setPixelColor(0, pixels.Color(LED_BRIGHTNESS, 0, 0)); // red
          pixels.show();
        }
        long_press = true;
      }
    }
    if(long_press) {
      turn_off = true;
    }
    else {
      if(turn_off)
        turn_off = false;
      else
        show_second = !show_second;
    }
  }

  // LED setting
  if(turn_off)
    pixels.setPixelColor(0, pixels.Color(LED_BRIGHTNESS, 0, 0)); // red
  else if(show_second) 
    pixels.setPixelColor(0, pixels.Color(0, 0, LED_BRIGHTNESS)); // blue
  else
    pixels.setPixelColor(0, pixels.Color(0, LED_BRIGHTNESS, 0)); // green
  pixels.show();

  tmtime.tm_hour %= 12;

  if(turn_off)
    return; // do nothing

  if(tmtime.tm_hour == 11 && tmtime.tm_min == 59 && tmtime.tm_sec >= 58) {
    ESP.restart(); // reset twice everyday
  }

  if(tmtime.tm_min >= 55 && tmtime.tm_sec >= tmtime.tm_min)
    return; // just wait for sec == 0

  if(!show_second) {
    tmtime.tm_sec = (tmtime.tm_min - 1 + 60) % 60; // place the second hand just behind the minute hand
  }
  n.v[0] = tmtime.tm_hour * 5 + tmtime.tm_min / 12;
  n.v[1] = tmtime.tm_min;
  n.v[2] = tmtime.tm_sec;
  if(n.v[1] > current.v[2] && current.v[2] > n.v[2]) {
    n.v[2] = n.v[2] + (current.v[2] - n.v[2]) / 15;
  }
  if(n.v[1] == n.v[2]) { // min and sec hands interfere
    n.v[2] = (n.v[2] + 4) % POSITION; // it takes around 5 sec for full turn
  }
  
  finished = setNumber(n);

  if(finished && tmtime.tm_hour != prevhour) {
    if(tmtime.tm_hour % 6 == 0)
      getNTP(); // acquire current time every 6 hours
    prevhour = tmtime.tm_hour;
  }
}
